@@ -125,7 +125,9 @@ static void parse_url(const gchar * url, gchar ** user, gchar ** pass, gchar **
 
 	if (!strncasecmp("http://", ptr, 7))
 		ptr += 7;
-	if ((h = strchr(ptr, '@')))
+	h = strchr(ptr, '@');
+	f = strchr(ptr, '/');
+	if (h != NULL && (!f || h < f))
 	{
 		*h = '\0';
 		p = strchr(ptr, ':');
@@ -148,7 +150,6 @@ static void parse_url(const gchar * url, gchar ** user, gchar ** pass, gchar **
 		h = ptr;
 	}
 	pt = strchr(ptr, ':');
-	f = strchr(ptr, '/');
 	if (pt != NULL && (f == NULL || pt < f))
 	{
 		*pt = '\0';
@@ -263,6 +264,8 @@ int mpg123_http_read(gpointer data, gint length)
 				}
 				g_free(meta_data);
 			}
+			if (!http_used())
+				http_wait_for_data(length - off);
 			cnt = min3(len, buffer_length - rd_index, http_used());
 		}
 		else if ((icy_metaint > 0) && (mpg123_cfg.cast_title_streaming))
@@ -320,7 +323,7 @@ gint mpg123_http_read_line(gchar * buf, gint size)
 	return i;
 }
 
-void *http_buffer_loop(void *arg)
+static void *http_buffer_loop(void *arg)
 {
 	gchar line[1024], *user, *pass, *host, *filename,
 	     *status, *url, *temp, *file;
@@ -552,7 +555,10 @@ void *http_buffer_loop(void *arg)
 		if(redirect)
 		{
 			if (output_file)
+			{
 				fclose(output_file);
+				output_file = NULL;
+			}
 			close(sock);
 			g_free(user);
 			g_free(pass);
@@ -632,13 +638,15 @@ void *http_buffer_loop(void *arg)
 		if (mpg123_cfg.use_udp_channel && udp_sock != 0)
 			if (udp_check_for_data(udp_sock) < 0)
 			{
-				g_message("closing udp");
 				close(udp_sock);
 				udp_sock = 0;
 			}
 	}
 	if (output_file)
+	{
 		fclose(output_file);
+		output_file = NULL;
+	}
 	close(sock);
 	if (udp_sock != 0)
 		close(udp_sock);
@@ -675,17 +683,19 @@ int mpg123_http_open(gchar * _url)
 	return 0;
 }
 
-gchar *mpg123_http_get_title(gchar * url)
+char *mpg123_http_get_title(char *url)
 {
 	if (icy_name)
 		return g_strdup(icy_name);
+	if (g_basename(url) && strlen(g_basename(url)) > 0)
+		return g_strdup(g_basename(url));
 	return g_strdup(url);
 }
 
 /* Start UDP Channel specific stuff */
 
 /* Find a good local udp port and bind udp_sock to it, return the port */
-static int udp_establish_listener (gint *sock)
+static int udp_establish_listener(int *sock)
 {
 	struct sockaddr_in sin;
 	socklen_t sinlen = sizeof (struct sockaddr_in);
@@ -801,8 +811,9 @@ static int udp_check_for_data(int sock)
 		else if (strstr(lines[i], "x-audiocast-streammsg") != NULL)
 		{
 			/*  mpg123_ip.set_info(title, -1, mpg123_bitrate * 1000, mpg123_frequency, mpg123_stereo); */
-			xmms_show_message(_("Message"), valptr, _("Ok"),
-					  FALSE, NULL, NULL);
+/*  			xmms_show_message(_("Message"), valptr, _("Ok"), */
+/*  					  FALSE, NULL, NULL); */
+			g_message("Stream_message: %s", valptr);
 		}
 
 #if 0
@@ -823,7 +834,7 @@ static int udp_check_for_data(int sock)
 			sprintf(obuf, "x-audiocast-ack: %ld \r\n", atol(valptr));
 			if (sendto(sock, obuf, strlen(obuf), 0, (struct sockaddr *) &from, fromlen) < 0)
 			{
-				g_log(NULL, G_LOG_LEVEL_CRITICAL,
+				g_log(NULL, G_LOG_LEVEL_WARNING,
 				      "udp_check_for_data(): Unable to send ack to server: %s", strerror(errno));
 			}
 #ifdef DEBUG_UDP
