@@ -872,6 +872,10 @@ static int III_dequantize_sample(real xr[SBLIMIT][SSLIMIT], int *scf,
 		{		/* short band? */
 			while (1)
 			{
+				/* HACK Prevent overflowing the xr buffer */
+				if (mc * 6 > &xr[SBLIMIT][SSLIMIT] - xrpnt)
+					return 1;
+					
 				for (; mc > 0; mc--)
 				{
 					*xrpnt = 0.0;
