@@ -1,6 +1,7 @@
 /*  XMMS - Cross-platform multimedia player
- *  Copyright (C) 1998-2001  Peter Alm, Mikael Alm, Olle Hallnas, Thomas Nilsson and 4Front Technologies
- *  Copyright (C) 1999-2001  Haavard Kvaalen
+ *  Copyright (C) 1998-2002  Peter Alm, Mikael Alm, Olle Hallnas,
+ *                           Thomas Nilsson and 4Front Technologies
+ *  Copyright (C) 1999-2002  Haavard Kvaalen
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -59,8 +60,6 @@ EqualizerPreset;
 
 static GList *equalizer_presets = NULL, *equalizer_auto_presets = NULL;
 
-#define EQUALIZERWIN_PRESETS_MENU_ENTRIES 20
-
 void equalizerwin_presets_menu_cb(gpointer cb_data, guint action, GtkWidget * w);
 GtkWidget * equalizerwin_create_conf_window(void);
 
@@ -69,6 +68,7 @@ enum
 	EQUALIZER_PRESETS_LOAD_PRESET, EQUALIZER_PRESETS_LOAD_AUTOPRESET,
 	EQUALIZER_PRESETS_LOAD_DEFAULT,	EQUALIZER_PRESETS_LOAD_ZERO,
 	EQUALIZER_PRESETS_LOAD_FROM_FILE, EQUALIZER_PRESETS_LOAD_FROM_WINAMPFILE,
+	EQUALIZER_PRESETS_IMPORT_WINAMPFILE,
 	EQUALIZER_PRESETS_SAVE_PRESET, EQUALIZER_PRESETS_SAVE_AUTOPRESET,
 	EQUALIZER_PRESETS_SAVE_DEFAULT,	EQUALIZER_PRESETS_SAVE_TO_FILE,
 	EQUALIZER_PRESETS_SAVE_TO_WINAMPFILE, EQUALIZER_PRESETS_DELETE_PRESET,
@@ -86,6 +86,8 @@ GtkItemFactoryEntry equalizerwin_presets_menu_entries[] =
 	{N_("/Load/-"), NULL, NULL, 0, "<Separator>"},
 	{N_("/Load/From file"), NULL, equalizerwin_presets_menu_cb, EQUALIZER_PRESETS_LOAD_FROM_FILE, "<Item>"},
 	{N_("/Load/From WinAMP EQF file"), NULL, equalizerwin_presets_menu_cb, EQUALIZER_PRESETS_LOAD_FROM_WINAMPFILE, "<Item>"},
+	{N_("/Import"), NULL, NULL, 0, "<Branch>"},
+	{N_("/Import/WinAMP Presets"), NULL, equalizerwin_presets_menu_cb, EQUALIZER_PRESETS_IMPORT_WINAMPFILE, "<Item>"},
 	{N_("/Save"), NULL, NULL, 0, "<Branch>"},
 	{N_("/Save/Preset"), NULL, equalizerwin_presets_menu_cb, EQUALIZER_PRESETS_SAVE_PRESET, "<Item>"},
 	{N_("/Save/Auto-load preset"), NULL, equalizerwin_presets_menu_cb, EQUALIZER_PRESETS_SAVE_AUTOPRESET, "<Item>"},
@@ -99,6 +101,10 @@ GtkItemFactoryEntry equalizerwin_presets_menu_entries[] =
 	{N_("/Configure Equalizer"), NULL, equalizerwin_presets_menu_cb, EQUALIZER_PRESETS_CONFIGURE, "<Item>"},
 };
 
+static gint equalizerwin_presets_menu_entries_num = 
+	sizeof(equalizerwin_presets_menu_entries) / 
+	sizeof(equalizerwin_presets_menu_entries[0]);
+
 void equalizerwin_set_shape_mask(void)
 {
 	if (cfg.show_wm_decorations)
@@ -131,7 +137,7 @@ void equalizerwin_set_doublesize(gboolean ds)
 	draw_equalizer_window(TRUE);
 }
 
-void equalizerwin_set_shade(gboolean shaded)
+void equalizerwin_set_shade_menu_cb(gboolean shaded)
 {
 	cfg.equalizer_shaded = shaded;
 
@@ -161,9 +167,17 @@ void equalizerwin_set_shade(gboolean shaded)
 	draw_equalizer_window(TRUE);
 }
 
-void equalizerwin_shade_cb(void)
+void equalizerwin_shade_toggle(void)
 {
-	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(gtk_item_factory_get_widget(mainwin_options_menu, "/Equalizer WindowShade Mode")), !cfg.equalizer_shaded);
+	equalizerwin_set_shade(!cfg.equalizer_shaded);
+}
+
+void equalizerwin_set_shade(gboolean shaded)
+{
+	GtkWidget *widget;
+        widget = gtk_item_factory_get_widget(mainwin_options_menu,
+					     "/Equalizer WindowShade Mode");
+	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(widget), shaded);
 }
 
 void equalizerwin_raise(void)
@@ -323,7 +337,7 @@ void equalizerwin_press(GtkWidget * widget, GdkEventButton * event, gpointer cal
 	}
 	else if (event->button == 1 && event->type == GDK_2BUTTON_PRESS && event->y < 14)
 	{
-		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(gtk_item_factory_get_widget(mainwin_options_menu, "/Equalizer WindowShade Mode")), !cfg.equalizer_shaded);
+		equalizerwin_set_shade(!cfg.equalizer_shaded);
 		if(dock_is_moving(equalizerwin))
 			dock_move_release(equalizerwin);
 	}
@@ -335,7 +349,7 @@ void equalizerwin_press(GtkWidget * widget, GdkEventButton * event, gpointer cal
 		 * Pop up the main menu a few pixels down to avoid
 		 * anything to be selected initially.
 		 */
-		util_item_factory_popup(mainwin_general_menu, event->x_root, event->y_root + 2, 3, GDK_CURRENT_TIME);
+		util_item_factory_popup(mainwin_general_menu, event->x_root, event->y_root + 2, 3, event->time);
 		grab = FALSE;
 	}
 	else
@@ -434,8 +448,13 @@ static gboolean equalizerwin_configure(GtkWidget * window, GdkEventConfigure *ev
 	if (!GTK_WIDGET_VISIBLE(window))
 		return FALSE;
 
-	gdk_window_get_root_origin(window->window, &cfg.equalizer_x, &cfg.equalizer_y);
-/*  	gdk_window_get_deskrelative_origin(window->window, &cfg.equalizer_x, &cfg.equalizer_y); */
+	if (cfg.show_wm_decorations)
+		gdk_window_get_root_origin(window->window,
+					   &cfg.equalizer_x, &cfg.equalizer_y);
+	else
+		gdk_window_get_deskrelative_origin(window->window,
+						   &cfg.equalizer_x,
+						   &cfg.equalizer_y);
 	return FALSE;
 }
 
@@ -598,7 +617,7 @@ static void equalizerwin_create_widgets(void)
 	equalizerwin_close = create_pbutton(&equalizerwin_wlist, equalizerwin_bg, equalizerwin_gc, 264, 3, 9, 9, 0, 116, 0, 125, equalizerwin_close_cb, SKIN_EQMAIN);
 	equalizerwin_close->pb_allow_draw = FALSE;
 
-	equalizerwin_shade = create_pbutton_ex(&equalizerwin_wlist, equalizerwin_bg, equalizerwin_gc, 254, 3, 9, 9, 254, 137, 1, 38, equalizerwin_shade_cb, SKIN_EQMAIN, SKIN_EQ_EX);
+	equalizerwin_shade = create_pbutton_ex(&equalizerwin_wlist, equalizerwin_bg, equalizerwin_gc, 254, 3, 9, 9, 254, 137, 1, 38, equalizerwin_shade_toggle, SKIN_EQMAIN, SKIN_EQ_EX);
 	equalizerwin_shade->pb_allow_draw = FALSE;
 
 	equalizerwin_graph = create_eqgraph(&equalizerwin_wlist, equalizerwin_bg, equalizerwin_gc, 86, 17);
@@ -649,6 +668,7 @@ static void equalizerwin_create_gtk(void)
 	gtk_widget_set_events(equalizerwin, GDK_FOCUS_CHANGE_MASK | GDK_BUTTON_MOTION_MASK | GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK);
 	gtk_widget_realize(equalizerwin);
 	hint_set_skip_winlist(equalizerwin);
+	util_set_cursor(equalizerwin);
 	if (!cfg.show_wm_decorations)
 		gdk_window_set_decorations(equalizerwin->window, 0);
 
@@ -686,7 +706,7 @@ void equalizerwin_create(void)
 	equalizerwin_presets_menu = gtk_item_factory_new(GTK_TYPE_MENU, "<Main>", equalizerwin_accel);
 	gtk_item_factory_set_translate_func(equalizerwin_presets_menu,
 					    util_menu_translate, NULL, NULL);
-	gtk_item_factory_create_items(equalizerwin_presets_menu, EQUALIZERWIN_PRESETS_MENU_ENTRIES, equalizerwin_presets_menu_entries, NULL);
+	gtk_item_factory_create_items(equalizerwin_presets_menu, equalizerwin_presets_menu_entries_num, equalizerwin_presets_menu_entries, NULL);
 	equalizer_presets = equalizerwin_read_presets("eq.preset");
 	equalizer_auto_presets = equalizerwin_read_presets("eq.auto_preset");
 
@@ -731,7 +751,7 @@ void equalizerwin_real_show(void)
 	draw_equalizer_window(TRUE);
 	cfg.equalizer_visible = TRUE;
 	tbutton_set_toggled(mainwin_eq, TRUE);
-	mainwin_set_always_on_top(cfg.always_on_top);
+	hint_set_always(cfg.always_on_top);
 	hint_set_sticky(cfg.sticky);
 	hint_set_skip_winlist(equalizerwin);
 }
@@ -849,6 +869,39 @@ static GList *equalizerwin_delete_preset(GList * list, gchar * name, gchar * fna
 	return list;
 }
 
+
+static GList *equalizerwin_import_winamp_eqf(FILE *file)
+{
+	gchar header[31];
+	gchar tmp[257];
+	gchar bands[11];
+	gint i=0;
+	GList *list = NULL;
+	EqualizerPreset *preset;
+
+	fread(header, 1, 31, file);
+	if (!strncmp(header, "Winamp EQ library file v1.1", 27))
+	{
+		while (fread(tmp, 1, 257, file)) {
+			preset = g_malloc(sizeof (EqualizerPreset));
+
+			fread(bands, 1, 11, file);
+
+			preset->name = g_strdup(tmp);
+			preset->preamp =  20.0 - ((bands[10] * 40.0) / 64);
+
+			for (i = 0; i < 10; i++)
+				preset->bands[i] = 20.0 - ((bands[i] * 40.0) / 64);
+
+			list = g_list_prepend(list, preset);
+		}
+        }
+
+	list = g_list_reverse(list);
+	return list;
+
+}
+
 static void equalizerwin_read_winamp_eqf(FILE *file)
 {
 	gchar header[31];
@@ -1029,6 +1082,26 @@ static void equalizerwin_load_filesel_ok(GtkWidget * w, GtkFileSelection * files
 	gtk_widget_destroy(GTK_WIDGET(filesel));
 }
 
+static void equalizerwin_import_winamp_filesel_ok(GtkWidget * w, GtkFileSelection * filesel)
+{
+	gchar *filename;
+	FILE *file;
+
+	if (util_filebrowser_is_dir(filesel))
+		return;
+
+	filename = gtk_file_selection_get_filename(filesel);
+
+	if ((file = fopen(filename, "rb")) != NULL)
+	{
+		equalizer_presets = g_list_concat(equalizer_presets,
+				    equalizerwin_import_winamp_eqf(file));
+		equalizerwin_write_preset_file(equalizer_presets, "eq.preset");
+	}
+
+	gtk_widget_destroy(GTK_WIDGET(filesel));
+}
+
 static void equalizerwin_load_winamp_filesel_ok(GtkWidget * w, GtkFileSelection * filesel)
 {
 	gchar *filename;
@@ -1256,6 +1329,22 @@ void equalizerwin_presets_menu_cb(gpointer cb_data, guint action, GtkWidget * w)
 			
 			break;
 		}
+		case EQUALIZER_PRESETS_IMPORT_WINAMPFILE:
+		{
+			static GtkWidget *import_winamp_filesel;
+
+			if (import_winamp_filesel != NULL)
+				break;
+
+			import_winamp_filesel = gtk_file_selection_new(_("Import equalizer presets"));
+			gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(import_winamp_filesel)->ok_button), "clicked", GTK_SIGNAL_FUNC(equalizerwin_import_winamp_filesel_ok), import_winamp_filesel);
+			gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(import_winamp_filesel)->cancel_button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(import_winamp_filesel));
+			gtk_signal_connect(GTK_OBJECT(import_winamp_filesel), "destroy", GTK_SIGNAL_FUNC(gtk_widget_destroyed), &import_winamp_filesel);
+			gtk_widget_show(import_winamp_filesel);
+			
+			break;
+		}
+	
 		case EQUALIZER_PRESETS_SAVE_PRESET:
 			if (!equalizerwin_save_window)
 				equalizerwin_create_list_window(equalizer_presets, _("Save preset"), &equalizerwin_save_window, GTK_SELECTION_SINGLE, &equalizerwin_save_entry, _("Ok"), _("Cancel"), equalizerwin_save_ok, equalizerwin_save_select);
