@@ -1,6 +1,7 @@
 /*  XMMS - Cross-platform multimedia player
- *  Copyright (C) 1998-2000  Peter Alm, Mikael Alm, Olle Hallnas, Thomas Nilsson and 4Front Technologies
- *  Copyright (C) 1999,2000  Hvard Kvlen
+ *  Copyright (C) 1998-2002  Peter Alm, Mikael Alm, Olle Hallnas,
+ *                           Thomas Nilsson and 4Front Technologies
+ *  Copyright (C) 1999-2002  Haavard Kvaalen
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -24,7 +25,8 @@
 
 GList *playlist = NULL;
 GList *shuffle_list = NULL;
-static gboolean playlist_get_info_scan_active = FALSE, playlist_get_info_going = FALSE;
+static gboolean playlist_get_info_scan_active = FALSE;
+static gboolean playlist_get_info_going = FALSE;
 static pthread_t playlist_get_info_thread;
 pthread_mutex_t playlist_mutex = PTHREAD_MUTEX_INITIALIZER;
 static PlaylistEntry *playlist_position;
@@ -98,9 +100,9 @@ void playlist_delete_node(GList *node, gboolean *set_info_text, gboolean *restar
 		}
 		playing_song = find_playlist_position_list();
 		if (g_list_next(playing_song))
-			playlist_position = g_list_next(playing_song)->data;
+			playlist_position = playing_song->next->data;
 		else if (g_list_previous(playing_song))
-			playlist_position = g_list_previous(playing_song)->data;
+			playlist_position = playing_song->prev->data;
 		else
 			playlist_position = NULL;
 		/* Make sure the entry did not disappear under us */
@@ -146,7 +148,7 @@ void playlist_delete_index(glong index)
 		if (playlist_position)
 			playlist_play();
 		else
-			mainwin_set_song_info(0, 0, 0);
+			mainwin_clear_song_info();
 	}
 	else if (set_info_text)
 		mainwin_set_info_text();
@@ -179,7 +181,7 @@ void playlist_delete_filenames(GList *filenames)
 		if (playlist_position)
 			playlist_play();
 		else
-			mainwin_set_song_info(0, 0, 0);
+			mainwin_clear_song_info();
 	}
 	else if (set_info_text)
 		mainwin_set_info_text();	
@@ -213,17 +215,19 @@ void playlist_delete(gboolean crop)
 		if (playlist_position)
 			playlist_play();
 		else
-			mainwin_set_song_info(0, 0, 0);
+			mainwin_clear_song_info();
 	}
 }
 
-static void __playlist_ins(gchar * filename, glong pos)
+static void __playlist_ins_with_info(char *filename, long pos, char* title, int len)
 {
 	PlaylistEntry *entry;
 
 	entry = g_malloc0(sizeof (PlaylistEntry));
 	entry->filename = g_strdup(filename);
-	entry->length = -1;
+	if (title)
+		entry->title = g_strdup(title);
+	entry->length = len;
 
 	pthread_mutex_lock(&playlist_mutex);
 	playlist = g_list_insert(playlist, entry, pos);
@@ -232,15 +236,30 @@ static void __playlist_ins(gchar * filename, glong pos)
 	playlist_get_info_scan_active = TRUE;
 }
 
+static void __playlist_ins(char * filename, long pos)
+{
+	__playlist_ins_with_info(filename, pos, NULL, -1);
+}
+
+static gboolean is_playlist_name(char *pathname)
+{
+	char *ext;
+
+	ext = strrchr (pathname, '.');
+	if (ext && ((!strcasecmp(ext, ".m3u") || !strcasecmp(ext, ".pls"))))
+		return TRUE;
+
+	return FALSE;
+}
+
+
 void playlist_ins(gchar * filename, glong pos)
 {
-	gchar *ext;
 	gboolean ispl = FALSE;
-	
-	ext = strrchr(filename, '.');
-	if (ext && ((!strcasecmp(ext, ".m3u") || !strcasecmp(ext, ".pls"))))
-		ispl = TRUE;
-	else if (!input_check_file(filename))
+
+	ispl = is_playlist_name (filename);
+
+	if (!ispl && !input_check_file(filename))
 	{
 		/*
 		 * Some files (typically produced by some cgi-scripts)
@@ -277,7 +296,6 @@ void playlist_ins(gchar * filename, glong pos)
 
 	}
 
-
 	if (ispl)
 		playlist_load_ins(filename, pos);
 	else
@@ -321,36 +339,45 @@ static guint playlist_ins_dir_real(gchar * path, glong pos)
 		return entries;
 	}
 
-	if ((dir = opendir(path)) != NULL)
+	if ((dir = opendir(path)) == NULL)
 	{
-		while ((dirent = readdir(dir)) != NULL)
-		{
-			if (dirent->d_name[0] != '.')
-			{
-				gchar *filename = g_strconcat(temp, dirent->d_name, NULL);
-				/* Why don't we follow symlinks? */
-				lstat(filename, &statbuf);
-				if (S_ISDIR(statbuf.st_mode))
-				{
-					guint i;
-					i = playlist_ins_dir_real(filename, pos);
-					entries += i;
-					if (pos >= 0)
-						pos += i;
-				}
-				else if (input_check_file(filename))
-				{
-					__playlist_ins(filename, pos);
-					entries++;
-					if (pos >= 0)
-						pos++;
-				}
-				g_free(filename);
-			}
-		}
-		closedir(dir);
+		g_free(temp);
+		return 0;
 	}
+
+	while ((dirent = readdir(dir)) != NULL)
+	{
+		char *filename;
+
+		if (dirent->d_name[0] == '.')
+			continue;
+
+		filename = g_strconcat(temp, dirent->d_name, NULL);
+
+		if (stat(filename, &statbuf) < 0)
+			goto next;
+
+		if (S_ISDIR(statbuf.st_mode))
+		{
+			guint i;
+			i = playlist_ins_dir_real(filename, pos);
+			entries += i;
+			if (pos >= 0)
+				pos += i;
+		}
+		else if (input_check_file(filename))
+		{
+			__playlist_ins(filename, pos);
+			entries++;
+			if (pos >= 0)
+				pos++;
+		}
+	next:
+		g_free(filename);
+	}
+	closedir(dir);
 	g_free(temp);
+
 	return entries;
 }
 
@@ -366,7 +393,7 @@ guint playlist_ins_dir(gchar * path, glong pos)
 
 guint playlist_ins_url_string(gchar * string, glong pos)
 {
-	gchar *temp, *ext;
+        gchar *temp;
 	struct stat statbuf;
 	gint i = 1, entries = 0;
 
@@ -389,8 +416,7 @@ guint playlist_ins_url_string(gchar * string, glong pos)
 			i = playlist_ins_dir(string, pos);
 		else
 		{
-			ext = strrchr(string, '.');
-			if (ext && (!strcasecmp(ext, ".m3u") || !strcasecmp(ext, ".pls")))
+			if (is_playlist_name (string))
 				i = playlist_load_ins(string, pos);
 			else
 			{
@@ -533,6 +559,7 @@ void playlist_next(void)
 		restart_playing = TRUE;
 	}
 
+	plist_pos_list = find_playlist_position_list();
 	if (g_list_next(plist_pos_list))
 		playlist_position = plist_pos_list->next->data;
 	else if (cfg.repeat)
@@ -587,8 +614,23 @@ void playlist_prev(void)
 		restart_playing = TRUE;
 	}
 	
+	plist_pos_list = find_playlist_position_list();
 	if (g_list_previous(plist_pos_list))
-		playlist_position = g_list_previous(plist_pos_list)->data;
+		playlist_position = plist_pos_list->prev->data;
+	else if (cfg.repeat)
+	{
+		GList *node;
+		playlist_position = NULL;
+		PL_UNLOCK();
+		playlist_generate_shuffle_list();
+		PL_LOCK();
+		if (cfg.shuffle)
+			node = g_list_last(shuffle_list);
+		else
+			node = g_list_last(playlist);
+		if (node)
+			playlist_position = node->data;
+	}
 	PL_UNLOCK();
 	playlist_check_pos_current();
 
@@ -654,7 +696,7 @@ void playlist_eof_reached(void)
 	if (cfg.no_playlist_advance)
 	{
 		PL_UNLOCK();
-		mainwin_set_song_info(0, 0, 0);
+		mainwin_clear_song_info();
 		if (cfg.repeat)
 			playlist_play();
 		return;
@@ -674,13 +716,13 @@ void playlist_eof_reached(void)
 		if (!cfg.repeat)
 		{
 			PL_UNLOCK();
-			mainwin_set_song_info(0, 0, 0);
+			mainwin_clear_song_info();
 			mainwin_set_info_text();
 			return;
 		}
 	}
 	else
-		playlist_position = g_list_next(plist_pos_list)->data;
+		playlist_position = plist_pos_list->next->data;
 	PL_UNLOCK();
 	playlist_check_pos_current();
 	playlist_play();
@@ -707,9 +749,9 @@ gint __get_playlist_length(void)
 	return 0;
 }
 
-gchar *playlist_get_info_text(void)
+char *playlist_get_info_text(void)
 {
-	gchar *text, *title, *tmp, *tmp2;
+	char *text, *title, *tmp, *numbers, *length;
 
 	PL_LOCK();
 	if (!playlist_position)
@@ -722,22 +764,37 @@ gchar *playlist_get_info_text(void)
 		title = playlist_position->title;
 	else
 		title = g_basename(playlist_position->filename);
-	
-	text = g_malloc(strlen(title) + 20);
-	if (playlist_position->length != -1)
-		text = g_strdup_printf("%d. %s (%d:%-2.2d)",
-				       __get_playlist_position() + 1,
-				       title, playlist_position->length / 60000,
-				       (playlist_position->length / 1000) % 60);
+
+	/*
+	 * If the user don't want numbers in the playlist, don't
+	 * display them in other parts of XMMS
+	 */
+
+	if (cfg.show_numbers_in_pl)
+		numbers = g_strdup_printf("%d. ", __get_playlist_position() + 1);
 	else
-		text = g_strdup_printf("%d. %s", __get_playlist_position() + 1,
-				       title);
+		numbers = g_strdup("");
+
+	if (playlist_position->length != -1)
+		length = g_strdup_printf(" (%d:%-2.2d)",
+					 playlist_position->length / 60000,
+					 (playlist_position->length / 1000) % 60);
+	else
+		length = g_strdup("");
+
+	text = g_strdup_printf("%s%s%s", numbers, title, length);
+	g_free(numbers);
+	g_free(length);
+
+	PL_UNLOCK();
+
 	if (cfg.convert_underscore)
 		while ((tmp = strchr(text, '_')) != NULL)
 			*tmp = ' ';
 	if (cfg.convert_twenty)
 		while ((tmp = strstr(text, "%20")) != NULL)
 		{
+			char *tmp2;
 			tmp2 = tmp + 3;
 			*(tmp++) = ' ';
 			while (*tmp2)
@@ -745,7 +802,6 @@ gchar *playlist_get_info_text(void)
 			*tmp = '\0';
 		}
 	
-	PL_UNLOCK();
 	return text;
 }
 
@@ -762,40 +818,55 @@ int playlist_get_current_length(void)
 
 gboolean playlist_save(gchar * filename)
 {
-	PlaylistEntry *entry;
 	GList *node;
 	FILE *file;
 	gchar *ext;
 	gboolean is_pls = FALSE;
 
-	if ((file = fopen(filename, "w")) != NULL)
+	if ((file = fopen(filename, "w")) == NULL)
+		return FALSE;
+
+	ext = strrchr(filename, '.');
+	if (ext && !strcasecmp(ext, ".pls"))
 	{
-		ext = strrchr(filename, '.');
-		if (ext)
-		{
-			if (!strcasecmp(ext, ".pls"))
-			{
-				is_pls = TRUE;
-				fprintf(file, "[playlist]\n");
-				fprintf(file, "NumberOfEntries=%d\n", get_playlist_length());
-			}
-		}
-		PL_LOCK();
-		node = playlist;
-		while (node)
-		{
-			entry = (PlaylistEntry *) node->data;
-			if (is_pls)
-				fprintf(file, "File%d=%s\n", g_list_position(playlist, node) + 1, entry->filename);
-			else
-				fprintf(file, "%s\n", entry->filename);
-			node = node->next;
-		}
-		PL_UNLOCK();
-		fclose(file);
-		return TRUE;
+		is_pls = TRUE;
+		fprintf(file, "[playlist]\n");
+		fprintf(file, "NumberOfEntries=%d\n", get_playlist_length());
 	}
-	return FALSE;
+
+	if (!is_pls)
+		fprintf(file, "#EXTM3U\n");
+
+	PL_LOCK();
+	node = playlist;
+	while (node)
+	{
+		PlaylistEntry *entry = node->data;
+		if (is_pls)
+			fprintf(file, "File%d=%s\n",
+				g_list_position(playlist, node) + 1,
+				entry->filename);
+		else
+		{
+			if (entry->title)
+			{
+				int seconds;
+				
+				if (entry->length > 0)
+					seconds = (entry->length) / 1000;
+				else
+					seconds = -1;
+
+				fprintf(file, "#EXTINF:%d,%s\n",
+					seconds, entry->title);
+			}
+			fprintf(file, "%s\n", entry->filename);
+		}
+		node = g_list_next(node);
+	}
+	PL_UNLOCK();
+	fclose(file);
+	return TRUE;
 }
 
 gboolean playlist_load(gchar * filename)
@@ -803,13 +874,13 @@ gboolean playlist_load(gchar * filename)
 	return playlist_load_ins(filename, -1);
 }
 
-static void playlist_load_ins_file(gchar * filename, gchar * playlist_name, glong pos)
+static void playlist_load_ins_file(char *filename, char *playlist_name, long pos, char *title, int len)
 {
 	gchar *temp, *path;
 
 	filename = g_strstrip(filename);
 
-	if(cfg.use_backslash_as_dir_delimiter)
+	if (cfg.use_backslash_as_dir_delimiter)
 	{
 		while ((temp = strchr(filename, '\\')) != NULL)
 		      *temp = '/';
@@ -823,43 +894,79 @@ static void playlist_load_ins_file(gchar * filename, gchar * playlist_name, glon
 			*temp = '\0';
 		else
 		{
-			__playlist_ins(filename, pos);
+			__playlist_ins_with_info(filename, pos, title, len);
 			return;
 		}
 		temp = g_strdup_printf("%s/%s", path, filename);
-		__playlist_ins(temp, pos);
+		__playlist_ins_with_info(temp, pos, title, len);
 		g_free(temp);
 		g_free(path);
 	}
 	else
-		__playlist_ins(filename, pos);
+		__playlist_ins_with_info(filename, pos, title, len);
+}
+
+static void parse_ext3u_info(char *info, char **title, int *length)
+{
+	char *str;
+	
+	*title = NULL;
+	*length = -1;
+
+	if (!info)
+		return;
+
+	g_return_if_fail(strlen(info) >= 8);
+	
+	*length = atoi(info + 8);
+	if (*length <= 0)
+		*length = -1;
+	else
+		*length *= 1000;
+	if ((str = strchr(info, ',')) != NULL)
+	{
+		*title = g_strdup(str + 1);
+		g_strstrip(*title);
+		if (strlen(*title) < 1)
+		{
+			g_free(*title);
+			*title = NULL;
+		}
+	}
 }
 
 guint playlist_load_ins(gchar * filename, glong pos)
 {
 	FILE *file;
-	gchar *line, *ext, key[10];
-	gint i, noe;
+	char *line, *ext;
 	guint entries = 0;
 	int linelen = 1024;
+	gboolean extm3u = FALSE;
+	char *ext_info = NULL, *ext_title = NULL;
+	int ext_len = -1;
 
 	ext = strrchr(filename, '.');
 	if (ext && !strcasecmp(ext, ".pls"))
 	{
-		if ((line = read_ini_string(filename, "playlist", "NumberOfEntries")) != NULL)
-		{
-			noe = atoi(line);
-			g_free(line);
-		}
-		else
-			return FALSE;
+		int noe, i;
+		char key[10];
+		
+		line = read_ini_string(filename, "playlist", "NumberOfEntries");
+		if (line == NULL)
+			return 0;
+
+		noe = atoi(line);
+		g_free(line);
+
 		for (i = 1; i <= noe; i++)
 		{
 			g_snprintf(key, 10, "File%d", i);
-			if ((line = read_ini_string(filename, "playlist", key)) != NULL)
+			line = read_ini_string(filename, "playlist", key);
+			if (line != NULL)
 			{
-				playlist_load_ins_file(line, filename, pos);
-				entries ++;
+				playlist_load_ins_file(line, filename, pos,
+						       NULL, -1);
+				entries++;
 				if (pos >= 0)
 					pos++;
 				g_free(line);
@@ -869,35 +976,78 @@ guint playlist_load_ins(gchar * filename, glong pos)
 		playlistwin_update_list();
 		return entries;
 	}
-	else
+
+	/*
+	 * Seems like an m3u.  Maybe we should do some sanity checking
+	 * here?  If someone accidentally selects something else, we
+	 * will try to read it.
+	 */
+
+	if ((file = fopen(filename, "r")) == NULL)
+		return 0;
+
+	line = g_malloc(linelen);
+	while (fgets(line, linelen, file))
 	{
-		if ((file = fopen(filename, "r")) != NULL)
+		while (strlen(line) == linelen - 1 &&
+		       line[strlen(line) - 1] == '\n')
 		{
-			line = g_malloc(linelen);
-			while (fgets(line, linelen, file))
-			{
-				while (strlen(line) == linelen - 1 && line[strlen(line) - 1] == '\n')
-				{
-					linelen += 1024;
-					line = (gchar *) g_realloc(line, linelen);
-					fgets(&line[strlen(line)], 1024, file);
-				}
-				while (line[strlen(line) - 1] == '\r' || line[strlen(line) - 1] == '\n')
-					line[strlen(line) - 1] = '\0';
-				if (line[0] == '#')
-					continue;
-				playlist_load_ins_file(line, filename, pos);
-				entries++;
-				if (pos >= 0)
-					pos++;
-			}
-			fclose(file);
-			playlist_generate_shuffle_list();
-			playlistwin_update_list();
-			return entries;
+			linelen += 1024;
+			line = g_realloc(line, linelen);
+			fgets(&line[strlen(line)], 1024, file);
 		}
+		while (line[strlen(line) - 1] == '\r' ||
+		       line[strlen(line) - 1] == '\n')
+			line[strlen(line) - 1] = '\0';
+		
+		if (!strncmp(line, "#EXTM3U", 8))
+		{
+			extm3u = TRUE;
+			continue;
+		}
+		
+		if (extm3u && !strncmp(line, "#EXTINF:", 8))
+		{
+			if (ext_info)
+				g_free(ext_info);
+			ext_info = g_strdup(line);
+			continue;
+		}
+		
+		if (line[0] == '#')
+		{
+			if (ext_info)
+			{
+				g_free(ext_info);
+				ext_info = NULL;
+			}
+			continue;
+		}
+		
+		if (extm3u)
+		{
+			if (cfg.read_pl_metadata)
+				parse_ext3u_info(ext_info, &ext_title, &ext_len);
+			g_free(ext_info);
+			ext_info = NULL;
+		}
+		
+		playlist_load_ins_file(line, filename, pos, ext_title, ext_len);
+		
+		g_free(ext_title);
+		ext_title = NULL;
+		ext_len = -1;
+		
+		entries++;
+		if (pos >= 0)
+			pos++;
 	}
-	return 0;
+	fclose(file);
+	g_free(line);
+	playlist_generate_shuffle_list();
+	playlistwin_update_list();
+
+	return entries;
 }
 
 GList *get_playlist(void)
@@ -1032,7 +1182,7 @@ gint playlist_get_songtime(gint pos)
 
 gint playlist_sort_by_title_cmpfunc(PlaylistEntry * a, PlaylistEntry * b)
 {
-	gchar *a_title = NULL, *b_title = NULL;
+	char *a_title, *b_title;
 
 	if (a->title)
 		a_title = a->title;
@@ -1066,7 +1216,7 @@ void playlist_sort_by_title(void)
 
 gint playlist_sort_by_filename_cmpfunc(PlaylistEntry * a, PlaylistEntry * b)
 {
-	gchar *a_filename = NULL, *b_filename = NULL;
+	char *a_filename, *b_filename;
 
 	if (strrchr(a->filename, '/'))
 		a_filename = strrchr(a->filename, '/') + 1;
@@ -1100,6 +1250,37 @@ void playlist_sort_by_path(void)
 	PL_UNLOCK();
 }
 
+gint playlist_sort_by_date_cmpfunc(PlaylistEntry * a, PlaylistEntry * b)
+{
+	struct stat buf;
+	time_t modtime;
+	
+	if(!lstat(a->filename, &buf))
+	{
+		modtime=buf.st_mtime;
+		if(!lstat(b->filename, &buf))
+		{
+			if(buf.st_mtime==modtime)
+				return 0;
+			else
+				return (buf.st_mtime-modtime)>0 ? -1 : 1;
+		}
+		else
+			return -1;
+	}
+	else if(!lstat(b->filename, &buf))
+		return 1;
+	else
+		return playlist_sort_by_filename_cmpfunc(a,b);
+}
+
+void playlist_sort_by_date(void)
+{
+	PL_LOCK();
+	playlist = g_list_sort(playlist, (GCompareFunc) playlist_sort_by_date_cmpfunc);
+	PL_UNLOCK();
+}
+
 static GList* playlist_sort_selected(GList *list, GCompareFunc cmpfunc)
 {
 	GList *list1, *list2;
@@ -1168,6 +1349,13 @@ void playlist_sort_selected_by_path(void)
 	PL_UNLOCK();
 }
 
+void playlist_sort_selected_by_date(void)
+{
+	PL_LOCK();
+	playlist = playlist_sort_selected(playlist, (GCompareFunc) playlist_sort_by_date_cmpfunc);
+	PL_UNLOCK();
+}
+
 void playlist_reverse(void)
 {
 	PL_LOCK();
@@ -1286,11 +1474,17 @@ void playlist_generate_shuffle_list(void)
 		mainwin_set_info_text();
 }
 
-void playlist_fileinfo(gint pos)
+void playlist_fileinfo(int pos)
 {
-	gchar *path = NULL;
+	char *path = NULL;
+	GList *node;
+	
 	PL_LOCK();
-	path = g_strdup(((PlaylistEntry *) g_list_nth(get_playlist(), pos)->data)->filename);
+	if ((node = g_list_nth(get_playlist(), pos)) != NULL)
+	{
+		PlaylistEntry *entry = node->data;
+		path = g_strdup(entry->filename);
+	}
 	PL_UNLOCK();
 	if (path)
 	{
@@ -1304,7 +1498,7 @@ void playlist_fileinfo_current(void)
 	gchar *path = NULL;
 	PL_LOCK();
 	
-	if (get_playlist())
+	if (get_playlist() && playlist_position)
 		path = g_strdup(playlist_position->filename);
 	PL_UNLOCK();
 	if (path)
@@ -1318,7 +1512,7 @@ static gboolean playlist_get_info_entry(PlaylistEntry *entry)
 {
 	/*
 	 * Caller need to hold playlist mutex.
-	 * Note that this function temporary drops the playlist mutex.
+	 * Note that this function temporarily drops the playlist mutex.
 	 * If it returns false, the entry might not be valid.
 	 */
 	gchar *temp_filename, *temp_title;
@@ -1373,10 +1567,7 @@ void *playlist_get_info_func(void *arg)
 				}
 				else if (entry->title || entry->length != -1)
 				{
-					int pos;
-					pos = g_list_position(get_playlist(), node);
-					if (playlistwin_item_visible(pos))
-						update_playlistwin = TRUE;
+					update_playlistwin = TRUE;
 					if (entry == playlist_position)
 						update_mainwin = TRUE;
 					break;
@@ -1417,10 +1608,7 @@ void *playlist_get_info_func(void *arg)
 				}
 				else if (entry->title || entry->length != -1)
 				{
-					int pos;
-					pos = g_list_position(get_playlist(), node);
-					if (playlistwin_item_visible(pos))
-						update_playlistwin = TRUE;
+					update_playlistwin = TRUE;
 					if (entry == playlist_position)
 						update_mainwin = TRUE;
 					found = TRUE;
@@ -1451,6 +1639,24 @@ void *playlist_get_info_func(void *arg)
 	pthread_exit(NULL);
 }
 
+void playlist_start_get_info_thread(void)
+{
+	playlist_get_info_going = TRUE;
+	pthread_create(&playlist_get_info_thread, NULL,
+		       playlist_get_info_func, NULL);
+}
+
+void playlist_stop_get_info_thread(void)
+{
+	playlist_get_info_going = FALSE;
+	pthread_join(playlist_get_info_thread, NULL);
+}
+
+void playlist_start_get_info_scan(void)
+{
+	playlist_get_info_scan_active = TRUE;
+}
+
 void playlist_remove_dead_files(void)
 {
 	/* FIXME? Does virtual directories work well? */
@@ -1514,35 +1720,153 @@ void playlist_remove_dead_files(void)
 	}
 }
 
-void playlist_start_get_info_thread(void)
+void playlist_get_total_time(gulong *total_time, gulong *selection_time, gboolean *total_more, gboolean *selection_more)
 {
-	playlist_get_info_going = TRUE;
-	pthread_create(&playlist_get_info_thread, NULL,
-		       playlist_get_info_func, NULL);
-}
-
-void playlist_stop_get_info_thread(void)
-{
-	playlist_get_info_going = FALSE;
-	pthread_join(playlist_get_info_thread, NULL);
-}
-
-void playlist_start_get_info_scan(void)
-{
-	playlist_get_info_scan_active = TRUE;
-}
-
-void playlist_selection_enqueue(GtkSelectionData *selection_data)
-{
-	gchar *str;
-
-	if (selection_data->type == GDK_SELECTION_TYPE_STRING && selection_data->length > 0)
+	GList *list;
+	PlaylistEntry *entry;
+	
+	*total_time = 0;
+	*selection_time = 0;
+	*total_more = FALSE;
+	*selection_more = FALSE;
+	
+	PL_LOCK();
+	list = get_playlist();
+	while (list)
 	{
-		str = g_malloc(selection_data->length + 1);
-		memcpy(str, selection_data->data, selection_data->length);
-		str[selection_data->length] = 0;
-		playlist_ins_url_string(str, -1);
-		playlistwin_update_list();
-		g_free(str);
+		entry = list->data;
+		if (entry->length != -1)
+			*total_time += entry->length / 1000;
+		else
+			*total_more = TRUE;
+		if (entry->selected)
+		{
+			if (entry->length != -1)
+				*selection_time += entry->length / 1000;
+			else
+				*selection_more = TRUE;
+		}
+		list = g_list_next(list);
 	}
+	PL_UNLOCK();
+}
+
+void playlist_select_all(gboolean set)
+{
+	GList *list;
+	
+	PL_LOCK();
+	list = get_playlist();
+	while (list)
+	{
+		PlaylistEntry *entry = list->data;
+		entry->selected = set;
+		list = list->next;
+	}
+	PL_UNLOCK();
+}
+
+void playlist_select_invert_all(void)
+{
+	GList *list;
+	
+	PL_LOCK();
+	list = get_playlist();
+	while (list)
+	{
+		PlaylistEntry *entry = list->data;
+		entry->selected = !entry->selected;
+		list = list->next;
+	}
+	PL_UNLOCK();
+}
+
+gboolean playlist_select_invert(int num)
+{
+	GList *list;
+	gboolean retv = FALSE;
+
+	PL_LOCK();
+	if ((list = g_list_nth(get_playlist(), num)) != NULL)
+	{
+		PlaylistEntry *entry = list->data;
+		entry->selected = !entry->selected;
+		retv = TRUE;
+	}
+	PL_UNLOCK();
+
+	return retv;
+}
+	
+
+void playlist_select_range(int min, int max, gboolean sel)
+{
+	GList *list;
+	int i;
+
+	if (min > max)
+	{
+		int tmp = min;
+		min = max;
+		max = tmp;
+	}
+
+	PL_LOCK();
+	list = g_list_nth(get_playlist(), min);
+	for (i = min; i <= max && list; i++)
+	{
+		PlaylistEntry *entry = list->data;
+		entry->selected = sel;
+		list = list->next;
+	}
+	PL_UNLOCK();
+}
+
+gboolean playlist_read_info_selection(void)
+{
+	GList *node;
+	gboolean retval = FALSE;
+
+	PL_LOCK();
+	for (node = get_playlist(); node; node = node->next)
+	{
+		PlaylistEntry *entry = node->data;
+		if (entry->selected)
+		{
+			retval = TRUE;
+			if (entry->title)
+				g_free(entry->title);
+			entry->title = NULL;
+			entry->length = -1;
+			if (!playlist_get_info_entry(entry))
+			{
+				if (g_list_index(get_playlist(), entry) == -1)
+					/* Entry disapeared while we
+					   looked it up.  Restart. */
+					node = get_playlist();
+			}
+		}
+	}
+	PL_UNLOCK();
+	playlistwin_update_list();
+	return retval;
+}
+
+void playlist_read_info(int pos)
+{
+	GList *node;
+
+	PL_LOCK();
+	if ((node = g_list_nth(get_playlist(), pos)) != NULL)
+	{
+		PlaylistEntry *entry = node->data;
+
+		if (entry->title)
+			g_free(entry->title);
+		entry->title = NULL;
+		entry->length = -1;
+		playlist_get_info_entry(entry);
+	}
+	PL_UNLOCK();
+	playlistwin_update_list();
 }
