@@ -1,6 +1,7 @@
 /*  XMMS - Cross-platform multimedia player
- *  Copyright (C) 1998-2001  Peter Alm, Mikael Alm, Olle Hallnas, Thomas Nilsson and 4Front Technologies
- *  Copyright (C) 1999-2001  Hvard Kvlen
+ *  Copyright (C) 1998-2002  Peter Alm, Mikael Alm, Olle Hallnas,
+ *                           Thomas Nilsson and 4Front Technologies
+ *  Copyright (C) 1999-2002  Haavard Kvaalen
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -21,7 +22,7 @@
 #include "libxmms/util.h"
 
 GtkWidget *playlistwin;
-GtkWidget *playlistwin_url_window = NULL, *playlistwin_dir_browser;
+static GtkWidget *playlistwin_url_window = NULL;
 GtkItemFactory *playlistwin_sort_menu, *playlistwin_sub_menu, *playlistwin_popup_menu;
 
 GdkPixmap *playlistwin_bg;
@@ -46,15 +47,14 @@ Vis *playlistwin_vis;
 
 static gboolean playlistwin_vis_enabled = FALSE;
 
-static gboolean selection_cleared = FALSE;
-
 enum
 {
 	ADD_URL, ADD_DIR, ADD_FILE,
 	SUB_MISC, SUB_ALL, SUB_CROP, SUB_SELECTED,
 	SEL_INV, SEL_ZERO, SEL_ALL,
 	MISC_SORT, MISC_FILEINFO, MISC_MISCOPTS,
-	PLIST_NEW, PLIST_SAVE, PLIST_LOAD
+	PLIST_NEW, PLIST_SAVE, PLIST_LOAD,
+	SEL_LOOKUP,
 };
 
 void playlistwin_sort_menu_callback(gpointer cb_data, guint action, GtkWidget * w);
@@ -63,46 +63,52 @@ void playlistwin_set_hints(void);
 
 enum
 {
-	PLAYLISTWIN_SORT_BYTITLE, PLAYLISTWIN_SORT_BYFILENAME, PLAYLISTWIN_SORT_BYPATH,
+	PLAYLISTWIN_SORT_BYTITLE, PLAYLISTWIN_SORT_BYFILENAME,
+	PLAYLISTWIN_SORT_BYPATH, PLAYLISTWIN_SORT_BYDATE,
 	PLAYLISTWIN_SORT_SEL_BYTITLE, PLAYLISTWIN_SORT_SEL_BYFILENAME,
-	PLAYLISTWIN_SORT_SEL_BYPATH, PLAYLISTWIN_SORT_RANDOMIZE, PLAYLISTWIN_SORT_REVERSE
+	PLAYLISTWIN_SORT_SEL_BYPATH, PLAYLISTWIN_SORT_SEL_BYDATE,
+	PLAYLISTWIN_SORT_RANDOMIZE, PLAYLISTWIN_SORT_REVERSE
 };
 
-#define PLAYLISTWIN_SORT_MENU_ENTRIES 11
-
 GtkItemFactoryEntry playlistwin_sort_menu_entries[] =
 {
 	{N_("/Sort List"), NULL, NULL, 0, "<Branch>"},
 	{N_("/Sort List/By Title"), NULL, playlistwin_sort_menu_callback, PLAYLISTWIN_SORT_BYTITLE, "<Item>"},
 	{N_("/Sort List/By Filename"), NULL, playlistwin_sort_menu_callback, PLAYLISTWIN_SORT_BYFILENAME, "<Item>"},
 	{N_("/Sort List/By Path + Filename"), NULL, playlistwin_sort_menu_callback, PLAYLISTWIN_SORT_BYPATH, "<Item>"},
+	{N_("/Sort List/By Date"), NULL, playlistwin_sort_menu_callback, PLAYLISTWIN_SORT_BYDATE, "<Item>"},
 	{N_("/Sort Selection"), NULL, NULL, 0, "<Branch>"},
 	{N_("/Sort Selection/By Title"), NULL, playlistwin_sort_menu_callback, PLAYLISTWIN_SORT_SEL_BYTITLE, "<Item>"},
 	{N_("/Sort Selection/By Filename"), NULL, playlistwin_sort_menu_callback, PLAYLISTWIN_SORT_SEL_BYFILENAME, "<Item>"},
 	{N_("/Sort Selection/By Path + Filename"), NULL, playlistwin_sort_menu_callback, PLAYLISTWIN_SORT_SEL_BYPATH, "<Item>"},
+	{N_("/Sort Selection/By Date"), NULL, playlistwin_sort_menu_callback, PLAYLISTWIN_SORT_SEL_BYDATE, "<Item>"},
 	{N_("/-"), NULL, NULL, 0, "<Separator>"},
 	{N_("/Randomize List"), NULL, playlistwin_sort_menu_callback, PLAYLISTWIN_SORT_RANDOMIZE, "<Item>"},
 	{N_("/Reverse List"), NULL, playlistwin_sort_menu_callback, PLAYLISTWIN_SORT_REVERSE, "<Item>"},
 };
 
+static const int playlistwin_sort_menu_entries_num = 
+	sizeof(playlistwin_sort_menu_entries) / 
+	sizeof(playlistwin_sort_menu_entries[0]);
+
 enum
 {
 	PLAYLISTWIN_REMOVE_DEAD_FILES, PLAYLISTWIN_PHYSICALLY_DELETE
 };
 
-#define PLAYLISTWIN_SUB_MENU_ENTRIES  2
-
 GtkItemFactoryEntry playlistwin_sub_menu_entries[] =
 {
 	{N_("/Remove Dead Files"), NULL, playlistwin_sub_menu_callback, PLAYLISTWIN_REMOVE_DEAD_FILES, "<Item>"},
 	{N_("/Physically Delete Files"), NULL, playlistwin_sub_menu_callback, PLAYLISTWIN_PHYSICALLY_DELETE, "<Item>"},
 };
 
-void playlistwin_popup_menu_callback(gpointer cb_data, guint action, GtkWidget * w); 
+static const int playlistwin_sub_menu_entries_num = 
+	sizeof(playlistwin_sub_menu_entries) / 
+	sizeof(playlistwin_sub_menu_entries[0]);
 
-#define PLAYLISTWIN_POPUP_MENU_ENTRIES  20
+void playlistwin_popup_menu_callback(gpointer cb_data, guint action, GtkWidget * w);
 
-GtkItemFactoryEntry playlistwin_popup_menu_entries[] = 
+GtkItemFactoryEntry playlistwin_popup_menu_entries[] =
 {
 	{N_("/View File Info"), NULL, playlistwin_popup_menu_callback, MISC_FILEINFO, "<Item>"},
 	{N_("/-"), NULL, NULL, 0, "<Separator>"},
@@ -121,6 +127,8 @@ GtkItemFactoryEntry playlistwin_popup_menu_entries[] =
 	{N_("/Selection/Select All"), NULL, playlistwin_popup_menu_callback, SEL_ALL, "<Item>"},
 	{N_("/Selection/Select None"), NULL, playlistwin_popup_menu_callback, SEL_ZERO, "<Item>"},
 	{N_("/Selection/Invert Selection"), NULL, playlistwin_popup_menu_callback, SEL_INV, "<Item>"},
+	{N_("/Selection/-"), NULL, NULL, 0, "<Separator>"},
+	{N_("/Selection/Read Extended Info"), NULL, playlistwin_popup_menu_callback, SEL_LOOKUP, "<Item>"},
 
 	{N_("/Sort"), NULL, NULL, 0, "<Item>"},
 
@@ -130,36 +138,20 @@ GtkItemFactoryEntry playlistwin_popup_menu_entries[] =
 	{N_("/Playlist/New List"), NULL, playlistwin_popup_menu_callback, PLIST_NEW, "<Item>"},
 };
 
+static const int playlistwin_popup_menu_entries_num = 
+	sizeof(playlistwin_popup_menu_entries) / 
+	sizeof(playlistwin_popup_menu_entries[0]);
+
 void playlistwin_draw_frame(void);
 
 static void playlistwin_update_info(void)
 {
-	GList *list;
-	PlaylistEntry *entry;
-	gchar *text, *sel_text, *tot_text;
-	gulong selection = 0, total = 0;
-	gboolean selection_more = FALSE, total_more = FALSE;
-
-	PL_LOCK();
-	list = get_playlist();
-	while (list)
-	{
-		entry = list->data;
-		if (entry->length != -1)
-			total += entry->length / 1000;
-		else
-			total_more = TRUE;
-		if (entry->selected)
-		{
-			if (entry->length != -1)
-				selection += entry->length / 1000;
-			else
-				selection_more = TRUE;
-		}
-		list = g_list_next(list);
-	}
-	PL_UNLOCK();
+	char *text, *sel_text, *tot_text;
+	gulong selection, total;
+	gboolean selection_more, total_more;
 
+	playlist_get_total_time(&total, &selection, &total_more, &selection_more);
+	
 	if (selection > 0 || (selection == 0 && !selection_more))
 	{
 		if (selection > 3600)
@@ -187,60 +179,46 @@ static void playlistwin_update_info(void)
 
 void playlistwin_update_sinfo(void)
 {
-	gchar *time = NULL, *pos, *title, *info, *tmp, *tmp2;
-	gint max_len, i;
-	PlaylistEntry *entry;
-	GList *playlist;
+	char *posstr, *timestr, *title, *info, *dots;
+	int pos, time, max_len;
 
-	PL_LOCK();
-	playlist = get_playlist();
+	pos = get_playlist_position();
+	title = playlist_get_songtitle(pos);
+	time = playlist_get_songtime(pos);
 
-	if (!playlist)
+	if (!title)
 	{
 		PL_UNLOCK();
 		textbox_set_text(playlistwin_sinfo, "");
 		return;
 	}
 
-	entry = g_list_nth(playlist, __get_playlist_position())->data;
-
-	title = entry->title;
-	if (!title)
-		title = g_basename(entry->filename);
-
-	max_len = (cfg.playlist_width - 35) / 5;
-
 	if (cfg.show_numbers_in_pl)
-		pos = g_strdup_printf("%d. ", __get_playlist_position() + 1);
+		posstr = g_strdup_printf("%d. ", pos + 1);
 	else
-		pos = g_strdup("");
+		posstr = g_strdup("");
 
-	if (entry->length != -1)
+	max_len = (cfg.playlist_width - 35) / 5 - strlen(posstr);
+
+	if (time != -1)
 	{
-		time = g_strdup_printf("%d:%-2.2d", entry->length / 60000, (entry->length / 1000) % 60);
-		max_len -= strlen(time);
-		info = g_strdup_printf("%s%-*.*s%s", pos, max_len - (int)strlen(pos), max_len - (int)strlen(pos), title, time);
-		g_free(time);
-
+		timestr = g_strdup_printf(" %d:%-2.2d", time / 60000,
+					  (time / 1000) % 60);
+		max_len -= strlen(timestr);
 	}
 	else
-		info = g_strdup_printf("%s%-*.*s", pos, max_len - (int)strlen(pos), max_len - (int)strlen(pos), title);
-	PL_UNLOCK();
-	g_free(pos);
-	if (strlen(title) > max_len - (time ? 1 : 0))
-	{
-		for (i = 0; i < 3; i++)
-			info[max_len - i - (time ? 2 : 1)] = '.';
-		if (time)
-			info[max_len - 1] = ' ';
-	}
+		timestr = g_strdup("");
 
 	if (cfg.convert_underscore)
-		while ((tmp = strchr(info, '_')) != NULL)
+	{
+		char *tmp;
+		while ((tmp = strchr(title, '_')) != NULL)
 			*tmp = ' ';
+	}
 	if (cfg.convert_twenty)
 	{
-		while ((tmp = strstr(info, "%20")) != NULL)
+		char *tmp, *tmp2;
+		while ((tmp = strstr(title, "%20")) != NULL)
 		{
 			tmp2 = tmp + 3;
 			*(tmp++) = ' ';
@@ -250,6 +228,20 @@ void playlistwin_update_sinfo(void)
 		}
 	}
 
+	if (strlen(title) > max_len)
+	{
+		max_len -= 3;
+		dots = "...";
+	}
+	else
+		dots = "";
+
+	info = g_strdup_printf("%s%-*.*s%s%s", posstr, max_len, max_len,
+			       title, dots, timestr);
+	g_free(posstr);
+	g_free(title);
+	g_free(timestr);
+
 	textbox_set_text(playlistwin_sinfo, info);
 	g_free(info);
 }
@@ -309,7 +301,7 @@ void playlistwin_create_mask(void)
 		gdk_bitmap_unref(tmp);
 }
 
-void playlistwin_set_shade(gboolean shaded)
+void playlistwin_set_shade_menu_cb(gboolean shaded)
 {
 	cfg.playlist_shaded = shaded;
 	if (!cfg.show_wm_decorations)
@@ -343,9 +335,17 @@ void playlistwin_set_shade(gboolean shaded)
 	draw_playlist_window(TRUE);
 }
 
-void playlistwin_shade_cb(void)
+void playlistwin_shade_toggle(void)
 {
-	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(gtk_item_factory_get_widget(mainwin_options_menu, "/Playlist WindowShade Mode")), !cfg.playlist_shaded);
+	playlistwin_set_shade(!cfg.playlist_shaded);
+}
+
+void playlistwin_set_shade(gboolean shaded)
+{
+	GtkWidget *widget;
+	widget = gtk_item_factory_get_widget(mainwin_options_menu,
+					    "/Playlist WindowShade Mode");
+	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(widget), shaded);
 }
 
 void playlistwin_raise(void)
@@ -357,10 +357,8 @@ void playlistwin_raise(void)
 void playlistwin_release(GtkWidget * widget, GdkEventButton * event, gpointer callback_data)
 {
 	if (event->button == 3)
-	{
-		g_message("3");
 		return;
-	}
+
 	gdk_pointer_ungrab(GDK_CURRENT_TIME);
 	gdk_flush();
 	if (playlistwin_resizing)
@@ -383,6 +381,12 @@ void playlistwin_release(GtkWidget * widget, GdkEventButton * event, gpointer ca
 	}
 }
 
+void playlistwin_scroll(int num)
+{
+	playlistwin_list->pl_first += num;
+	playlistwin_update_list();
+}
+
 void playlistwin_scroll_up_pushed(void)
 {
 	playlistwin_list->pl_first -= 3;
@@ -397,16 +401,7 @@ void playlistwin_scroll_down_pushed(void)
 
 void playlistwin_select_all(void)
 {
-	GList *node;
-
-	PL_LOCK();
-	node = get_playlist();
-	while (node)
-	{
-		((PlaylistEntry *) node->data)->selected = TRUE;
-		node = node->next;
-	}
-	PL_UNLOCK();
+	playlist_select_all(TRUE);
 	playlistwin_list->pl_prev_selected=0;
 	playlistwin_list->pl_prev_min=0;
 	playlistwin_list->pl_prev_max=get_playlist_length()-1;
@@ -415,16 +410,7 @@ void playlistwin_select_all(void)
 
 void playlistwin_select_none(void)
 {
-	GList *node;
-
-	PL_LOCK();
-	node = get_playlist();
-	while (node)
-	{
-		((PlaylistEntry *) node->data)->selected = FALSE;
-		node = node->next;
-	}
-	PL_UNLOCK();
+	playlist_select_all(FALSE);
 	playlistwin_list->pl_prev_selected=-1;
 	playlistwin_list->pl_prev_min=-1;
 	playlistwin_update_list();
@@ -432,16 +418,7 @@ void playlistwin_select_none(void)
 
 void playlistwin_inverse_selection(void)
 {
-	GList *node;
-
-	PL_LOCK();
-	node = get_playlist();
-	while (node)
-	{
-		((PlaylistEntry *) node->data)->selected = !((PlaylistEntry *) node->data)->selected;
-		node = node->next;
-	}
-	PL_UNLOCK();
+	playlist_select_invert_all();
 	playlistwin_list->pl_prev_selected=-1;
 	playlistwin_list->pl_prev_min=-1;
 	playlistwin_update_list();
@@ -596,15 +573,21 @@ void playlistwin_add_dir_handler(gchar * dir)
 
 void playlistwin_show_dirbrowser(void)
 {
-	if (!playlistwin_dir_browser)
-	{
-		playlistwin_dir_browser = xmms_create_dir_browser(_("Select directory to add:"), cfg.filesel_path, GTK_SELECTION_EXTENDED, playlistwin_add_dir_handler);
-		gtk_signal_connect(GTK_OBJECT(playlistwin_dir_browser), "destroy", GTK_SIGNAL_FUNC(gtk_widget_destroyed), &playlistwin_dir_browser);
-		gtk_window_set_transient_for(GTK_WINDOW(playlistwin_dir_browser), GTK_WINDOW(playlistwin));
-	}
-	if (!GTK_WIDGET_VISIBLE(playlistwin_dir_browser))
-		gtk_widget_show(playlistwin_dir_browser);
+	static GtkWidget *dir_browser;
 
+	if (dir_browser)
+		return;
+
+	dir_browser = xmms_create_dir_browser(_("Select directory to add:"),
+					      cfg.filesel_path,
+					      GTK_SELECTION_EXTENDED,
+					      playlistwin_add_dir_handler);
+	gtk_signal_connect(GTK_OBJECT(dir_browser),
+			   "destroy", GTK_SIGNAL_FUNC(gtk_widget_destroyed),
+			   &dir_browser);
+	gtk_window_set_transient_for(GTK_WINDOW(dir_browser),
+				     GTK_WINDOW(playlistwin));
+	gtk_widget_show(dir_browser);
 }
 
 static void playlistwin_fileinfo(void)
@@ -641,6 +624,9 @@ static void playlistwin_set_sensitive_sortmenu(void)
 	w = gtk_item_factory_get_widget(playlistwin_sort_menu,
 					"/Sort Selection/By Path + Filename");
 	gtk_widget_set_sensitive(w, set);
+	w = gtk_item_factory_get_widget(playlistwin_sort_menu,
+					"/Sort Selection/By Date");
+	gtk_widget_set_sensitive(w, set);
 }
 
 void playlistwin_save_filesel_ok(GtkWidget * w, GtkWidget * filesel)
@@ -683,7 +669,7 @@ void playlistwin_load_filesel_ok(GtkWidget * w, GtkWidget * filesel)
 	if (filename && *filename)
 	{
 		playlist_clear();
-		mainwin_set_song_info(0, 0, 0);
+		mainwin_clear_song_info();
 		mainwin_set_info_text();
 
 		playlist_load(filename);
@@ -779,7 +765,7 @@ void playlistwin_popup_handler(gint item)
 			break;
 		case SUB_ALL:
 			playlist_clear();
-			mainwin_set_song_info(0, 0, 0);
+			mainwin_clear_song_info();
 			mainwin_set_info_text();
 			break;
 		case SUB_CROP:
@@ -819,7 +805,7 @@ void playlistwin_popup_handler(gint item)
 		/* Playlist button */
 		case PLIST_NEW:
 			playlist_clear();
-			mainwin_set_song_info(0, 0, 0);
+			mainwin_clear_song_info();
 			mainwin_set_info_text();
 			break;
 		case PLIST_SAVE:
@@ -956,11 +942,11 @@ void playlistwin_press(GtkWidget * widget, GdkEventButton * event, gpointer call
 		else
 			cfg.timer_mode = TIMER_ELAPSED;
 	}
-	else if (event->button == 2 && (event->type == GDK_BUTTON_PRESS || event->type == GDK_2BUTTON_PRESS))
+	else if (event->button == 2 && (event->type == GDK_BUTTON_PRESS) &&
+		inside_widget(event->x, event->y, playlistwin_list))
 	{
-		if(event->type == GDK_2BUTTON_PRESS)
-			selection_cleared = TRUE;
-		gtk_selection_convert(widget, GDK_SELECTION_PRIMARY, GDK_TARGET_STRING, GDK_CURRENT_TIME);
+		gtk_selection_convert(widget, GDK_SELECTION_PRIMARY,
+				      GDK_TARGET_STRING, event->time);
 	}
 	else if (cfg.playlist_width >= 350 && REGION_R(223, 151, 26, 10))
 	{
@@ -977,7 +963,7 @@ void playlistwin_press(GtkWidget * widget, GdkEventButton * event, gpointer call
 			GdkModifierType modmask;
 
 			gdk_window_get_pointer(NULL, &mx, &my, &modmask);
-			util_item_factory_popup(mainwin_vis_menu, mx, my, 3, GDK_CURRENT_TIME);
+			util_item_factory_popup(mainwin_vis_menu, mx, my, 3, event->time);
 			grab = FALSE;
 		}
 	}
@@ -992,7 +978,7 @@ void playlistwin_press(GtkWidget * widget, GdkEventButton * event, gpointer call
 	else if (event->button == 1 && event->type == GDK_2BUTTON_PRESS &&
 		 !inside_sensitive_widgets(event->x, event->y) && event->y < 14)
 	{
-		playlistwin_shade_cb();
+		playlistwin_set_shade(!cfg.playlist_shaded);
 		if(dock_is_moving(playlistwin))
 			dock_move_release(playlistwin);
 	}
@@ -1011,7 +997,7 @@ void playlistwin_press(GtkWidget * widget, GdkEventButton * event, gpointer call
 		 * Pop up the main menu a few pixels down to avoid
 		 * anything to be selected initially.
 		 */
-		util_item_factory_popup(mainwin_general_menu, event->x_root, event->y_root + 2, 3, GDK_CURRENT_TIME);
+		util_item_factory_popup(mainwin_general_menu, event->x_root, event->y_root + 2, 3, event->time);
 		grab = FALSE;
 	}
 	else if (event->button == 3 &&
@@ -1031,13 +1017,13 @@ void playlistwin_press(GtkWidget * widget, GdkEventButton * event, gpointer call
 						  GINT_TO_POINTER(pos), NULL,
 						  event->x_root,
 						  event->y_root + 5,
-						  3, GDK_CURRENT_TIME);
+						  3, event->time);
 		grab = FALSE;
 	}
-	else if (event->button == 4)
-		playlistwin_scroll_up_pushed();
-	else if(event->button == 5)
-		playlistwin_scroll_down_pushed();
+	else if (event->button == 4) /* Scrollwheel up */
+		playlistwin_scroll(-3);
+	else if(event->button == 5) /* Scrollwheel down */
+		playlistwin_scroll(3);
 	else
 	{
 		handle_press_cb(playlistwin_wlist, widget, event);
@@ -1071,15 +1057,19 @@ static gboolean playlistwin_configure(GtkWidget * window, GdkEventConfigure *eve
 	if (!GTK_WIDGET_VISIBLE(window))
 		return FALSE;
 	
-	gdk_window_get_root_origin(window->window, &cfg.playlist_x, &cfg.playlist_y);
-/*  	gdk_window_get_deskrelative_origin(window->window, &cfg.playlist_x, &cfg.playlist_y); */
-
 	if (cfg.show_wm_decorations)
 	{
 		if (event->width != cfg.playlist_width ||
 		    event->height != cfg.playlist_height)
 			playlistwin_resize(event->width, event->height);
+		gdk_window_get_root_origin(window->window,
+					   &cfg.playlist_x, &cfg.playlist_y);
 	}
+	else
+		gdk_window_get_deskrelative_origin(window->window,
+						   &cfg.playlist_x,
+						   &cfg.playlist_y);
+
 	return FALSE;
 }
 
@@ -1195,14 +1185,13 @@ void playlistwin_physically_delete(void)
 	PL_LOCK();
 	for (node = get_playlist(); node != 0; node = g_list_next(node))
 	{
+		PlaylistEntry *entry = node->data;
 		/*
 		 * We do not want to try to unlink url's
 		 */
-		if (((PlaylistEntry *) node->data)->selected &&
-		    !strstr(((PlaylistEntry *) node->data)->filename, ":/"))
-		{
-			selected_list = g_list_prepend(selected_list, g_strdup(((PlaylistEntry *) node->data)->filename));
-		}
+		if (entry->selected && !strstr(entry->filename, ":/"))
+			selected_list = g_list_prepend(selected_list,
+						       g_strdup(entry->filename));
 	}
 	PL_UNLOCK();
 	selected_list = g_list_reverse(selected_list);
@@ -1250,8 +1239,6 @@ void playlistwin_physically_delete(void)
 
 gboolean playlistwin_keypress(GtkWidget * w, GdkEventKey * event, gpointer data)
 {
-	GList *list;
-	PlaylistEntry *entry;
 	guint keyval;
 	gboolean refresh = TRUE;
 
@@ -1281,17 +1268,7 @@ gboolean playlistwin_keypress(GtkWidget * w, GdkEventKey * event, gpointer data)
 				}
 			}
 			else
-			{
-				PL_LOCK();
-				list = get_playlist();
-				while (list)
-				{
-					entry = (PlaylistEntry *) list->data;
-					entry->selected = FALSE;
-					list = list->next;
-				}
-				PL_UNLOCK();
-			}
+				playlist_select_all(FALSE);
 
 			if (playlistwin_list->pl_prev_selected == -1 ||
 			    (!playlistwin_item_visible(playlistwin_list->pl_prev_selected) &&
@@ -1311,7 +1288,7 @@ gboolean playlistwin_keypress(GtkWidget * w, GdkEventKey * event, gpointer data)
 
 				playlistwin_list->pl_first = MIN(playlistwin_list->pl_first, playlistwin_list->pl_prev_max);
 				playlistwin_list->pl_first = MAX(playlistwin_list->pl_first, playlistwin_list->pl_prev_max - playlistwin_list->pl_num_visible + 1);
-				playlist_list_select_range(playlistwin_list->pl_prev_min, playlistwin_list->pl_prev_max, TRUE);
+				playlist_select_range(playlistwin_list->pl_prev_min, playlistwin_list->pl_prev_max, TRUE);
 				break;
 			}
 			else if (keyval == GDK_Up)
@@ -1328,18 +1305,14 @@ gboolean playlistwin_keypress(GtkWidget * w, GdkEventKey * event, gpointer data)
 
 			if (event->state & GDK_MOD1_MASK)
 				break;
-			if ((list = g_list_nth(get_playlist(), playlistwin_list->pl_prev_selected)) != NULL)
-			{
-				entry = list->data;
-				entry->selected = TRUE;
-				playlistwin_list->pl_prev_min = -1;
-			}
+			playlist_select_range(playlistwin_list->pl_prev_selected, playlistwin_list->pl_prev_selected, TRUE);
+			playlistwin_list->pl_prev_min = -1;
 			break;
 		case GDK_Page_Up:
-			playlistwin_scroll_up_pushed();
+			playlistwin_scroll(-playlistwin_list->pl_num_visible);
 			break;
 		case GDK_Page_Down:
-			playlistwin_scroll_down_pushed();
+			playlistwin_scroll(playlistwin_list->pl_num_visible);
 			break;
 		case GDK_Home:
 			playlistwin_list->pl_first = 0;
@@ -1511,6 +1484,10 @@ void playlistwin_sort_menu_callback(gpointer cb_data, guint action, GtkWidget *
 			playlist_sort_by_path();
 			playlistwin_update_list();
 			break;
+		case PLAYLISTWIN_SORT_BYDATE:
+			playlist_sort_by_date();
+			playlistwin_update_list();
+			break;
 		case PLAYLISTWIN_SORT_SEL_BYTITLE:
 			playlist_sort_selected_by_title();
 			playlistwin_update_list();
@@ -1523,6 +1500,10 @@ void playlistwin_sort_menu_callback(gpointer cb_data, guint action, GtkWidget *
 			playlist_sort_selected_by_path();
 			playlistwin_update_list();
 			break;
+		case PLAYLISTWIN_SORT_SEL_BYDATE:
+			playlist_sort_selected_by_date();
+			playlistwin_update_list();
+			break;
 		case PLAYLISTWIN_SORT_REVERSE:
 			playlist_reverse();
 			playlistwin_update_list();
@@ -1632,13 +1613,14 @@ static void playlistwin_create_widgets(void)
 	if (!cfg.playlist_shaded)
 		hide_widget(playlistwin_sinfo);
 	if (cfg.playlist_shaded)
-		playlistwin_shade = create_pbutton(&playlistwin_wlist, playlistwin_bg, playlistwin_gc, cfg.playlist_width - 21, 3, 9, 9, 128, 45, 150, 42, playlistwin_shade_cb, SKIN_PLEDIT);
+		playlistwin_shade = create_pbutton(&playlistwin_wlist, playlistwin_bg, playlistwin_gc, cfg.playlist_width - 21, 3, 9, 9, 128, 45, 150, 42, playlistwin_shade_toggle, SKIN_PLEDIT);
 	else
-		playlistwin_shade = create_pbutton(&playlistwin_wlist, playlistwin_bg, playlistwin_gc, cfg.playlist_width - 21, 3, 9, 9, 157, 3, 62, 42, playlistwin_shade_cb, SKIN_PLEDIT);
+		playlistwin_shade = create_pbutton(&playlistwin_wlist, playlistwin_bg, playlistwin_gc, cfg.playlist_width - 21, 3, 9, 9, 157, 3, 62, 42, playlistwin_shade_toggle, SKIN_PLEDIT);
 	playlistwin_shade->pb_allow_draw = FALSE;
 	playlistwin_close = create_pbutton(&playlistwin_wlist, playlistwin_bg, playlistwin_gc, cfg.playlist_width - 11, 3, 9, 9, cfg.playlist_shaded ? 138 : 167, cfg.playlist_shaded ? 45 : 3, 52, 42, playlistwin_close_cb, SKIN_PLEDIT);
 	playlistwin_close->pb_allow_draw = FALSE;
 	playlistwin_list = create_playlist_list(&playlistwin_wlist, playlistwin_bg, playlistwin_gc, 12, 20, cfg.playlist_width - 31, cfg.playlist_height - 58);
+	playlist_list_set_font(cfg.playlist_font);
 	playlistwin_slider = create_playlistslider(&playlistwin_wlist, playlistwin_bg, playlistwin_gc, cfg.playlist_width - 15, 20, cfg.playlist_height - 58, playlistwin_list);
 	playlistwin_time_min = create_textbox(&playlistwin_wlist, playlistwin_bg, playlistwin_gc, cfg.playlist_width - 82, cfg.playlist_height - 15, 15, FALSE, SKIN_TEXT);
 	playlistwin_time_sec = create_textbox(&playlistwin_wlist, playlistwin_bg, playlistwin_gc, cfg.playlist_width - 64, cfg.playlist_height - 15, 10, FALSE, SKIN_TEXT);
@@ -1659,11 +1641,9 @@ static void playlistwin_create_widgets(void)
 
 static void selection_received(GtkWidget *widget, GtkSelectionData *selection_data, gpointer data)
 {
-	if(selection_cleared)
-		playlist_clear();
-	playlist_selection_enqueue(selection_data);
-	if(selection_cleared)
-		playlist_play();
+	if (selection_data->type == GDK_SELECTION_TYPE_STRING &&
+	    selection_data->length > 0)
+		playlist_add_url_string(selection_data->data);
 }
 
 static void playlistwin_create_gtk(void)
@@ -1685,6 +1665,7 @@ static void playlistwin_create_gtk(void)
 	gtk_widget_realize(playlistwin);
 	hint_set_skip_winlist(playlistwin);
 	playlistwin_set_hints();
+	util_set_cursor(playlistwin);
 
 	gtk_signal_connect(GTK_OBJECT(playlistwin), "delete_event", GTK_SIGNAL_FUNC(playlistwin_delete), NULL);
 	gtk_signal_connect(GTK_OBJECT(playlistwin), "button_press_event", GTK_SIGNAL_FUNC(playlistwin_press), NULL);
@@ -1715,11 +1696,13 @@ void playlistwin_create(void)
 	gtk_item_factory_set_translate_func(playlistwin_sort_menu,
 					    util_menu_translate, NULL, NULL);
 	gtk_item_factory_create_items(GTK_ITEM_FACTORY(playlistwin_sort_menu),
-				      PLAYLISTWIN_SORT_MENU_ENTRIES,
+				      playlistwin_sort_menu_entries_num,
 				      playlistwin_sort_menu_entries, NULL);
 	playlistwin_sub_menu = gtk_item_factory_new(GTK_TYPE_MENU, "<Main>", NULL);
+	gtk_item_factory_set_translate_func(playlistwin_sub_menu,
+					    util_menu_translate, NULL, NULL);
 	gtk_item_factory_create_items(GTK_ITEM_FACTORY(playlistwin_sub_menu),
-				      PLAYLISTWIN_SUB_MENU_ENTRIES,
+				      playlistwin_sub_menu_entries_num,
 				      playlistwin_sub_menu_entries, NULL);
 	playlistwin_bg = gdk_pixmap_new(NULL, cfg.playlist_width,
 					cfg.playlist_height,
@@ -1730,7 +1713,7 @@ void playlistwin_create(void)
 	gtk_item_factory_set_translate_func(playlistwin_popup_menu,
 					    util_menu_translate, NULL, NULL);
 	gtk_item_factory_create_items(GTK_ITEM_FACTORY(playlistwin_popup_menu),
-				      PLAYLISTWIN_POPUP_MENU_ENTRIES,
+				      playlistwin_popup_menu_entries_num,
 				      playlistwin_popup_menu_entries, NULL);
 
 	item = gtk_item_factory_get_widget(playlistwin_popup_menu, "/Sort");
@@ -1781,7 +1764,7 @@ void playlistwin_real_show(void)
 	cfg.playlist_visible = TRUE;
 	playlistwin_set_toprow(0);
 	playlist_check_pos_current();
-	mainwin_set_always_on_top(cfg.always_on_top);
+	hint_set_always(cfg.always_on_top);
 	hint_set_sticky(cfg.sticky);
 	hint_set_skip_winlist(playlistwin);
 }
@@ -1802,6 +1785,9 @@ void playlistwin_popup_menu_callback(gpointer cb_data, guint action, GtkWidget *
 			if (pos != -1)
 				playlist_fileinfo(pos);
 			break;
+		case SEL_LOOKUP:
+			playlist_read_info_selection();
+			break;
 		default:
 			playlistwin_popup_handler(action);
 	}
